//
// <copyright file="SecUtil.cs" company="Microsoft">
//     Copyright (c) Microsoft Corporation.  All rights reserved.
// </copyright>
//
using System;
using System.Globalization;
using System.Web.Hosting;
using System.Collections;
using System.Collections.Specialized;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using System.Configuration.Provider;
using System.Configuration;
using System.Text.RegularExpressions;
using System.Xml;
using System.Net;
using System.Diagnostics;
using System.Security;
using Microsoft.ServiceHosting.ServiceRuntime;
using Microsoft.Samples.ServiceHosting.StorageClient;


namespace Microsoft.Samples.ServiceHosting.ClouderaCMS
{

    internal static class SecUtility
    {

        internal const int Infinite = Int32.MaxValue;

        internal static bool ValidateParameter(ref string param, bool checkForNull, bool checkIfEmpty, bool checkForCommas, int maxSize)
        {
            if (param == null)
            {
                return !checkForNull;
            }

            param = param.Trim();
            if ((checkIfEmpty && param.Length < 1) ||
                 (maxSize > 0 && param.Length > maxSize) ||
                 (checkForCommas && param.Contains(",")))
            {
                return false;
            }

            return true;
        }

        internal static void CheckParameter(ref string param, bool checkForNull, bool checkIfEmpty, bool checkForCommas, int maxSize, string paramName)
        {
            if (param == null)
            {
                if (checkForNull)
                {
                    throw new ArgumentNullException(paramName);
                }

                return;
            }

            param = param.Trim();
            if (checkIfEmpty && param.Length < 1)
            {
                throw new ArgumentException(string.Format(CultureInfo.InstalledUICulture, "The parameter '{0}' must not be empty.", paramName), paramName);
            }

            if (maxSize > 0 && param.Length > maxSize)
            {
                throw new ArgumentException(string.Format(CultureInfo.InstalledUICulture, "The parameter '{0}' is too long: it must not exceed {1} chars in length.", paramName, maxSize.ToString(CultureInfo.InvariantCulture)), paramName);
            }

            if (checkForCommas && param.Contains(","))
            {
                throw new ArgumentException(string.Format(CultureInfo.InstalledUICulture, "The parameter '{0}' must not contain commas.", paramName), paramName);
            }
        }

        internal static void CheckArrayParameter(ref string[] param, bool checkForNull, bool checkIfEmpty, bool checkForCommas, int maxSize, string paramName)
        {
            if (param == null)
            {
                throw new ArgumentNullException(paramName);
            }

            if (param.Length < 1)
            {
                throw new ArgumentException(string.Format(CultureInfo.InstalledUICulture, "The array parameter '{0}' should not be empty.", paramName), paramName);
            }

            Hashtable values = new Hashtable(param.Length);
            for (int i = param.Length - 1; i >= 0; i--)
            {
                SecUtility.CheckParameter(ref param[i], checkForNull, checkIfEmpty, checkForCommas, maxSize,
                    paramName + "[ " + i.ToString(CultureInfo.InvariantCulture) + " ]");
                if (values.Contains(param[i]))
                {
                    throw new ArgumentException(string.Format(CultureInfo.InstalledUICulture, "The array '{0}' should not contain duplicate values.", paramName), paramName);
                }
                else
                {
                    values.Add(param[i], param[i]);
                }
            }
        }

        internal static void SetUtcTime(DateTime value, out DateTime res)
        {
            res = TableStorageConstants.MinSupportedDateTime;
            if ((value.Kind == DateTimeKind.Local && value.ToUniversalTime() < TableStorageConstants.MinSupportedDateTime) ||
                 value < TableStorageConstants.MinSupportedDateTime)
            {
                throw new ArgumentException("Invalid time value!");
            }
            if (value.Kind == DateTimeKind.Local)
            {
                res = value.ToUniversalTime();
            }
            else
            {
                res = value;
            }
        }

        internal static bool IsValidTableName(string name)
        {
            if (string.IsNullOrEmpty(name))
            {
                return false;
            }
            Regex reg = new Regex(StorageClient.StorageHttpConstants.RegularExpressionStrings.ValidTableNameRegex);
            if (reg.IsMatch(name))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        internal static bool IsValidContainerName(string name)
        {
            if (string.IsNullOrEmpty(name))
            {
                return false;
            }            
            Regex reg = new Regex(StorageClient.StorageHttpConstants.RegularExpressionStrings.ValidContainerNameRegex);
            if (reg.IsMatch(name))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        // the table storage system currently does not support the StartsWith() operation in 
        // queries. As a result we transform s.StartsWith(substring) into s.CompareTo(substring) > 0 && 
        // s.CompareTo(NextComparisonString(substring)) < 0
        // we assume that comparison on the service side is as ordinal comparison
        internal static string NextComparisonString(string s)
        {
            if (string.IsNullOrEmpty(s))
            {
                throw new ArgumentException("The string argument must not be null or empty!");
            }
            string ret;
            char last = s[s.Length - 1];
            if ((int)last + 1 > (int)char.MaxValue)
            {
                throw new ArgumentException("Cannot convert the string.");
            }
            // don't use "as" because we want to have an explicit exception here if something goes wrong
            last = (char)((int)last + 1);
            ret = s.Substring(0, s.Length - 1) + last;
            return ret;
        }

        /*
        internal const char KeySeparator = '$';
        internal static readonly string[] replace = { "&", "$", "\\" };
        internal static readonly string[] with = { "&1", "&2", "&3" };
         */

        internal const char KeySeparator = 'a';
        internal static readonly string[] replace = { "b", "a", "\\" };
        internal static readonly string[] with = { "b1", "b2", "b3" };

        // remove '$' and '\' characters
        internal static string Escape(string s)
        {
            if (string.IsNullOrEmpty(s)) {
                return s;
            }
            string ret = s;
            for (int i = 0; i < replace.Length; i++) {
                ret = ret.Replace(replace[i], with[i]);
            } 
            return ret;
        }

        internal static string UnEscape(string s)
        {
            if (string.IsNullOrEmpty(s))
            {
                return s;
            }
            string ret = s;
            for (int i = with.Length; i > 0; i--)
            {
                ret = ret.Replace(with[i], replace[i]);
            }
            return ret;
        }

        internal static string CombineToKey(string s1, string s2)
        {
            return Escape(s1) + KeySeparator + Escape(s2);
        }

        internal static string EscapedFirst(string s)
        {
            return Escape(s) + KeySeparator;
        }

        internal static string GetFirstFromKey(string key) {
            Debug.Assert(key.IndexOf(KeySeparator) != -1);
            string first = key.Substring(0, key.IndexOf(KeySeparator));
            return UnEscape(first);
        }

        internal static string GetSecondFromKey(string key)
        {
            Debug.Assert(key.IndexOf(KeySeparator) != -1);
            string second = key.Substring(key.IndexOf(KeySeparator) + 1);
            return UnEscape(second);
        }

        internal static void CheckAllowInsecureEndpoints(bool allowInsecureRemoteEndpoints, StorageAccountInfo info)
        {
            if (info == null)
            {
                throw new ArgumentNullException("The parameter specifying the StorageAccountInfo object must not be null.");
            }
            if (allowInsecureRemoteEndpoints)
            {
                return;
            }
            if (info.BaseUri == null)
            {
                throw new SecurityException("allowInsecureRemoteEndpoints is set to false (default setting) but the endpoint URL seems to be empty." +
                                            "Please configure the provider to use an https enpoint for the storage endpoint or " +
                                            "explicitly set the configuration option allowInsecureRemoteEndpoints to true.");
            }
            if (string.IsNullOrEmpty(info.BaseUri.Scheme) || info.BaseUri.Scheme.ToUpper(CultureInfo.InvariantCulture) == "HTTPS")
            {
                return;
            }
            if (info.BaseUri.Host == "127.0.0.1")
            {
                return;
            }
            throw new SecurityException("The provider is configured with allowInsecureRemoteEndpoints set to false (default setting) but the endpoint for " +
                                        "the storage system does not seem to be an https or local endpoint. As the local host only the IP address " +
                                        "127.0.0.1 is allowed, not the string localhost." +
                                        "Please configure the provider to use an https enpoint for the storage endpoint or " +
                                        "explicitly set the configuration option allowInsecureRemoteEndpoints to true.");
        }
    }

    internal static class Constants
    {
        internal const int MaxTableUsernameLength = 256;
        internal const int MaxTableApplicationNameLength = 256;
    }

    /// <summary>
    /// We are using this retry policies for only one purpose: the ASP providers often read data from the server, process it 
    /// locally and then write the result back to the server. The problem is that the row that has been read might have changed 
    /// between the read and write operation. This retry policy is used to retry the whole process in this case.
    /// </summary>
    internal static class ProviderRetryPolicies
    {
        /// <summary>
        /// Policy that does no retries i.e., it just invokes <paramref name="action"/> exactly once
        /// </summary>
        /// <param name="action">The action to retry</param>
        /// <returns>The return value of <paramref name="action"/></returns>
        internal static void NoRetry(Action action)
        {
            action();
        }

        /// <summary>
        /// Policy that retries a specified number of times with a specified fixed time interval between retries
        /// </summary>
        /// <param name="numberOfRetries">The number of times to retry. Should be a non-negative number</param>
        /// <param name="intervalBetweenRetries">The time interval between retries. Use TimeSpan.Zero to specify immediate
        /// retries</param>
        /// <returns></returns>
        /// <remarks>When <paramref name="numberOfRetries"/> is 0 and <paramref name="intervalBetweenRetries"/> is
        /// TimeSpan.Zero this policy is equivalent to the NoRetry policy</remarks>
        internal static RetryPolicy RetryN(int numberOfRetries, TimeSpan intervalBetweenRetries)
        {
            return new RetryPolicy((Action action) =>
            {
                RetryNImpl(action, numberOfRetries, intervalBetweenRetries);
            }
            );
        }

        private static void RetryNImpl(Action action, int numberOfRetries, TimeSpan intervalBetweenRetries)
        {
            do
            {
                try
                {
                    action();
                    break;
                }
                catch (InvalidOperationException e)
                {
                    HttpStatusCode status;
                    // precondition failed is the status code returned by the server to indicate that the etag is wrong
                    if (TableStorageHelpers.EvaluateException(e, out status))
                    {
                        if (status == HttpStatusCode.PreconditionFailed)
                        {
                            if (numberOfRetries == 0)
                                throw;
                            if (intervalBetweenRetries > TimeSpan.Zero)
                                System.Threading.Thread.Sleep(intervalBetweenRetries);
                        }
                        else
                        {
                            throw;
                        }
                    }
                    else
                    {
                        throw;
                    }
                }
            }
            while (numberOfRetries-- > 0);
        }
    }


    internal enum EventKind
    {
        Critical,
        Error,
        Warning,
        Information,
        Verbose
    };

    internal static class Log
    {
        internal static void WriteToLog(string logName, string fmt, params object[] args)
        {
            RoleManager.WriteToLog(logName, String.Format(CultureInfo.InvariantCulture, fmt, args));
        }
        internal static void Write(EventKind eventKind, string message, params object[] args)
        {
            if (RoleManager.IsRoleManagerRunning)
            {
                switch (eventKind)
                {
                    case EventKind.Error:
                        WriteToLog("Error", message, args);
                        break;
                    case EventKind.Critical:
                        WriteToLog("Critical", message, args);
                        break;
                    case EventKind.Warning:
                        WriteToLog("Warning", message, args);
                        break;
                    case EventKind.Information:
                        WriteToLog("Information", message, args);
                        break;
                    case EventKind.Verbose:
                        WriteToLog("Verbose", message, args);
                        break;
                }
            }
            else
            {
                switch (eventKind)
                {
                    case EventKind.Error:
                    case EventKind.Critical:
                        Trace.TraceError(message, args);
                        break;
                    case EventKind.Warning:
                        Trace.TraceWarning(message, args);
                        break;
                    case EventKind.Information:
                    case EventKind.Verbose:
                        Trace.TraceInformation(message, args);
                        break;
                }
            }
        }
    }

}
